{% extends "default/module.html.jinja2" %}

{% block favicon %}
    <link rel="icon" type="image/svg+xml" href="data:image/svg+xml,{% filter urlencode %}{% include "cobib_book.svg" %}{% endfilter %}"/>
{% endblock %}

{% block nav_title %}
    <br/>
    <img src="data:image/svg+xml,{% filter urlencode %}{% include "cobib_logo.svg" %}{% endfilter %}" />
{% endblock %}

<!-- vim: set ft=html: -->
