{% extends "default/index.html.jinja2" %}

{% block title %}coBib Documentation{% endblock %}

{% block favicon %}
    <link rel="icon" type="image/svg+xml" href="data:image/svg+xml,{% filter urlencode %}{% include "cobib_book.svg" %}{% endfilter %}"/>
{% endblock %}

{% block logo %}
    <a id="pdoc-logo" title="coBib" href="https://gitlab.com/cobib/cobib"></a>
        <img src="data:image/svg+xml,{% filter urlencode %}{% include "cobib_logo.svg" %}{% endfilter %}" alt="coBib logo" width="600"/>
    </a>
{% endblock %}

<!-- vim: set ft=html: -->
